@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2009-2020 Vadim Mikhailov
+ * Copyright (c) 2009-2022 Vadim Mikhailov
  *
  * Utility to turn USB port power on/off
  * for USB hubs that support per-port power switching.
@@ -47,6 +47,10 @@ int snprintf(char * __restrict __str, size_t __size, const char * __restrict __f
 #include <time.h>   /* for nanosleep */
 #endif
 
+#ifdef __gnu_linux__
+#include <fcntl.h> /* for open() / O_WRONLY */
+#endif
+
 /* cross-platform sleep function */
 
 void sleep_ms(int milliseconds)
@@ -221,6 +225,18 @@ static int opt_wait   = 20; /* wait before repeating in ms */
 static int opt_exact  = 0;  /* exact location match - disable USB3 duality handling */
 static int opt_reset  = 0;  /* reset hub after operation(s) */
 static int opt_force  = 0;  /* force operation even on unsupported hubs */
+static int opt_nodesc = 0;  /* skip querying device description */
+#ifdef __gnu_linux__
+static int opt_nosysfs = 0; /* don't use the Linux sysfs port disable interface, even if available */
+#endif
+
+
+static const char short_options[] =
+    "l:L:n:a:p:d:r:w:s:hvefRN"
+#ifdef __gnu_linux__
+    "S"
+#endif
+;
 
 static const struct option long_options[] = {
     { "location", required_argument, NULL, 'l' },
@@ -234,6 +250,10 @@ static const struct option long_options[] = {
     { "wait",     required_argument, NULL, 'w' },
     { "exact",    no_argument,       NULL, 'e' },
     { "force",    no_argument,       NULL, 'f' },
+    { "nodesc",   no_argument,       NULL, 'N' },
+#ifdef __gnu_linux__
+    { "nosysfs",  no_argument,       NULL, 'S' },
+#endif
     { "reset",    no_argument,       NULL, 'R' },
     { "version",  no_argument,       NULL, 'v' },
     { "help",     no_argument,       NULL, 'h' },
@@ -259,6 +279,10 @@ static int print_usage()
         "--repeat,   -r - repeat power off count [%d] (some devices need it to turn off).\n"
         "--exact,    -e - exact location (no USB3 duality handling).\n"
         "--force,    -f - force operation even on unsupported hubs.\n"
+        "--nodesc,   -N - do not query device description (helpful for unresponsive devices).\n"
+#ifdef __gnu_linux__
+        "--nosysfs,  -S - do not use the Linux sysfs port disable interface.\n"
+#endif
         "--reset,    -R - reset hub after each power-on action, causing all devices to reassociate.\n"
         "--wait,     -w - wait before repeat power off [%d ms].\n"
         "--version,  -v - print program version.\n"
@@ -504,6 +528,113 @@ static int get_port_status(struct libusb_device_handle *devh, int port)
 }
 
 
+#ifdef __gnu_linux__
+/*
+ * Try to use the Linux sysfs interface to power a port off/on.
+ * Returns 0 on success.
+ */
+
+static int set_port_status_linux(struct libusb_device_handle *devh, struct hub_info *hub, int port, int on)
+{
+    int configuration = 0;
+    char disable_path[PATH_MAX];
+
+    int rc = libusb_get_configuration(devh, &configuration);
+    if (rc < 0) {
+        return rc;
+    }
+
+    /*
+     * The "disable" sysfs interface is available only starting with kernel version 6.0.
+     * For earlier kernel versions the open() call will fail and we fall back to using libusb.
+     */
+    snprintf(disable_path, PATH_MAX,
+        "/sys/bus/usb/devices/%s:%d.0/%s-port%i/disable",
+        hub->location, configuration, hub->location, port
+    );
+
+    int disable_fd = open(disable_path, O_WRONLY);
+    if (disable_fd >= 0) {
+        rc = write(disable_fd, on ? "0" : "1", 1);
+        close(disable_fd);
+    }
+
+    if (disable_fd < 0 || rc < 0) {
+        /*
+         * ENOENT is the expected error when running on Linux kernel < 6.0 where
+         * sysfs disable interface does not exist yet - no need to report anything in this case.
+         * If the file exists but another error occurs it is most likely a permission issue.
+         * Print an error message mostly geared towards setting up udev.
+         */
+        if (errno != ENOENT) {
+            fprintf(stderr,
+                "Failed to set port status by writing to %s (%s).\n"
+                "Follow https://git.io/JIB2Z to make sure that udev is set up correctly.\n"
+                "Falling back to libusb based port control.\n"
+                "Use -S to skip trying the sysfs interface and printing this message.\n",
+                disable_path, strerror(errno)
+            );
+        }
+
+        return -1;
+    }
+
+    return 0;
+}
+#endif
+
+
+/*
+ * Use a control transfer via libusb to turn a port off/on.
+ * Returns >= 0 on success.
+ */
+
+static int set_port_status_libusb(struct libusb_device_handle *devh, int port, int on)
+{
+    int rc = 0;
+    int request = on ? LIBUSB_REQUEST_SET_FEATURE
+                     : LIBUSB_REQUEST_CLEAR_FEATURE;
+    int repeat = on ? 1 : opt_repeat;
+
+    while (repeat-- > 0) {
+        rc = libusb_control_transfer(devh,
+            LIBUSB_REQUEST_TYPE_CLASS | LIBUSB_RECIPIENT_OTHER,
+            request, USB_PORT_FEAT_POWER,
+            port, NULL, 0, USB_CTRL_GET_TIMEOUT
+        );
+        if (rc < 0) {
+            perror("Failed to control port power!\n");
+        }
+        if (repeat > 0) {
+            sleep_ms(opt_wait);
+        }
+    }
+
+    return rc;
+}
+
+
+/*
+ * Try different methods to power a port off/on.
+ * Return >= 0 on success.
+ */
+
+static int set_port_status(struct libusb_device_handle *devh, struct hub_info *hub, int port, int on)
+{
+#ifdef __gnu_linux__
+    if (!opt_nosysfs) {
+        if (set_port_status_linux(devh, hub, port, on) == 0) {
+            return 0;
+        }
+    }
+#else
+    (void)hub;
+#endif
+
+    return set_port_status_libusb(devh, port, on);
+}
+
+
 /*
  * Get USB device descriptor strings and summary description.
  *
@@ -535,20 +666,22 @@ static int get_device_description(struct libusb_device * dev, struct descriptor_
     id_product = libusb_le16_to_cpu(desc.idProduct);
     rc = libusb_open(dev, &devh);
     if (rc == 0) {
-        if (desc.iManufacturer) {
-            libusb_get_string_descriptor_ascii(devh,
-                desc.iManufacturer, (unsigned char*)ds->vendor, sizeof(ds->vendor));
-            rtrim(ds->vendor);
-        }
-        if (desc.iProduct) {
-            libusb_get_string_descriptor_ascii(devh,
-                desc.iProduct, (unsigned char*)ds->product, sizeof(ds->product));
-            rtrim(ds->product);
-        }
-        if (desc.iSerialNumber) {
-            libusb_get_string_descriptor_ascii(devh,
-                desc.iSerialNumber, (unsigned char*)ds->serial, sizeof(ds->serial));
-            rtrim(ds->serial);
+        if (!opt_nodesc) {
+            if (desc.iManufacturer) {
+                rc = libusb_get_string_descriptor_ascii(devh,
+                    desc.iManufacturer, (unsigned char*)ds->vendor, sizeof(ds->vendor));
+                rtrim(ds->vendor);
+            }
+            if (rc >= 0 && desc.iProduct) {
+                rc = libusb_get_string_descriptor_ascii(devh,
+                    desc.iProduct, (unsigned char*)ds->product, sizeof(ds->product));
+                rtrim(ds->product);
+            }
+            if (rc >= 0 && desc.iSerialNumber) {
+                rc = libusb_get_string_descriptor_ascii(devh,
+                    desc.iSerialNumber, (unsigned char*)ds->serial, sizeof(ds->serial));
+                rtrim(ds->serial);
+            }
         }
         if (desc.bDeviceClass == LIBUSB_CLASS_HUB) {
             struct hub_info info;
@@ -899,8 +1032,7 @@ int main(int argc, char *argv[])
     int option_index = 0;
 
     for (;;) {
-        c = getopt_long(argc, argv, "l:L:n:a:p:d:r:w:s:hvefR",
-            long_options, &option_index);
+        c = getopt_long(argc, argv, short_options, long_options, &option_index);
         if (c == -1)
             break;  /* no more options left */
         switch (c) {
@@ -914,16 +1046,16 @@ int main(int argc, char *argv[])
             printf("\n");
             break;
         case 'l':
-            strncpy(opt_location, optarg, sizeof(opt_location));
+            snprintf(opt_location, sizeof(opt_location), "%s", optarg);
             break;
         case 'L':
             opt_level = atoi(optarg);
             break;
         case 'n':
-            strncpy(opt_vendor, optarg, sizeof(opt_vendor));
+            snprintf(opt_vendor, sizeof(opt_vendor), "%s", optarg);
             break;
         case 's':
-            strncpy(opt_search, optarg, sizeof(opt_search));
+            snprintf(opt_search, sizeof(opt_search), "%s", optarg);
             break;
         case 'p':
             if (!strcasecmp(optarg, "all")) { /* all ports is the default */
@@ -956,6 +1088,14 @@ int main(int argc, char *argv[])
         case 'f':
             opt_force = 1;
             break;
+        case 'N':
+            opt_nodesc = 1;
+            break;
+#ifdef __gnu_linux__
+        case 'S':
+            opt_nosysfs = 1;
+            break;
+#endif
         case 'e':
             opt_exact = 1;
             break;
@@ -1033,7 +1173,7 @@ int main(int argc, char *argv[])
             continue;
         if (k == 1 && opt_action == POWER_KEEP)
             continue;
-        // if toggle requested, do it only once when `k == 0`
+        /* if toggle requested, do it only once when `k == 0` */
         if (k == 1 && opt_action == POWER_TOGGLE)
             continue;
         int i;
@@ -1052,45 +1192,29 @@ int main(int argc, char *argv[])
             if (rc == 0) {
                 /* will operate on these ports */
                 int ports = ((1 << hubs[i].nports) - 1) & opt_ports;
-                int request = (k == 0) ? LIBUSB_REQUEST_CLEAR_FEATURE
-                                       : LIBUSB_REQUEST_SET_FEATURE;
+                int should_be_on = k;
+
                 int port;
                 for (port=1; port <= hubs[i].nports; port++) {
                     if ((1 << (port-1)) & ports) {
                         int port_status = get_port_status(devh, port);
                         int power_mask = hubs[i].super_speed ? USB_SS_PORT_STAT_POWER
                                                              : USB_PORT_STAT_POWER;
-                        int powered_on = port_status & power_mask;
+                        int is_on = (port_status & power_mask) != 0;
+
                         if (opt_action == POWER_TOGGLE) {
-                            request = powered_on ? LIBUSB_REQUEST_CLEAR_FEATURE
-                                                 : LIBUSB_REQUEST_SET_FEATURE;
+                            should_be_on = !is_on;
                         }
-                        if (k == 0 && !powered_on && opt_action != POWER_TOGGLE)
-                            continue;
-                        if (k == 1 && powered_on)
-                            continue;
-                        int repeat = powered_on ? opt_repeat : 1;
-                        while (repeat-- > 0) {
-                            rc = libusb_control_transfer(devh,
-                                LIBUSB_REQUEST_TYPE_CLASS | LIBUSB_RECIPIENT_OTHER,
-                                request, USB_PORT_FEAT_POWER,
-                                port, NULL, 0, USB_CTRL_GET_TIMEOUT
-                            );
-                            if (rc < 0) {
-                                perror("Failed to control port power!\n");
-                            }
-                            if (repeat > 0) {
-                                sleep_ms(opt_wait);
-                            }
+
+                        if (is_on != should_be_on) {
+                            rc = set_port_status(devh, &hubs[i], port, should_be_on);
                         }
                     }
                 }
                 /* USB3 hubs need extra delay to actually turn off: */
                 if (k==0 && hubs[i].super_speed)
                     sleep_ms(150);
-                printf("Sent power %s request\n",
-                    request == LIBUSB_REQUEST_CLEAR_FEATURE ? "off" : "on"
-                );
+                printf("Sent power %s request\n", should_be_on ? "on" : "off");
                 printf("New status for hub %s [%s]\n",
                     hubs[i].location, hubs[i].ds.description
                 );
